package Source;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JProgressBar;

public class Forest implements java.io.Serializable {
    
    long id;
    int forestArea;                     // in ha
    int tph;                            // trees/ha
    int treeCount;
    double minTreeDistance;             // in m
    int clusterCount;
    double clusterFactor;               // factor acting on distance, eg <1 for clustering, >1 for decustering
    double clusterRadius;               // in m
    int clusterizationRatio;            // percentage 0 to 100
    boolean randomCluster;
    String folding;
    int repetition;
    private Grid grid;
    private Grid clusterGrid;
    int forestSide;                     // in m
    PointList[] clusterList;
    
    public Forest(int forestArea, int treeDensity, double minTreeDistance, int clusterCount, double clusterFactor, double clusterRadius, int clusterizationRatio, boolean randomCluster, String folding, int repetition) {
        //store the parameters
        this.id = System.currentTimeMillis();
        this.forestArea = forestArea;
        this.tph = treeDensity;
        this.treeCount = forestArea * treeDensity;
        this.minTreeDistance = minTreeDistance;
        this.clusterCount = clusterCount;
        this.clusterFactor = clusterFactor;
        this.clusterRadius = clusterRadius;
        this.clusterizationRatio = clusterizationRatio;
        this.randomCluster = randomCluster;
        this.folding = folding;
        this.repetition = repetition;
        this.forestSide = (int) Math.floor(Math.sqrt(forestArea * 10000));

        //generate empty grid structures
        if (folding.equals("Torus")) {
            this.grid = new TorusGrid(forestSide);
            this.clusterGrid = new TorusGrid(forestSide);
            
        } else if (folding.equals("Sphere")) {
            this.grid = new SphereGrid(forestSide);
            this.clusterGrid = new SphereGrid(forestSide);
        } else {
            //default
            this.grid = new TorusGrid(forestSide);
            this.clusterGrid = new TorusGrid(forestSide);
        }
    }
    
    public void generate(final JProgressBar progressBar) {
        this.id = System.currentTimeMillis();
        grid.generateRandom(treeCount, minTreeDistance, progressBar);
        
        if (clusterFactor != 0) {
            //generate cluster centers 
            progressBar.setString("Generating cluster centers..");
            progressBar.setValue(0);
            clusterGrid.generateRandom(clusterCount, clusterRadius * 2, progressBar);
            PointList clusterCenters = clusterGrid.getPointList();
            clusterList = new PointList[clusterCount];
            for (int i = 0; i < clusterCount; i++) {
                clusterList[i] = new PointList();
                progressBar.setValue((100 * i) / forestSide);
                progressBar.paint(progressBar.getGraphics());
            }

            //assign cluster # to each point
            progressBar.setString("Identifying clusters..");
            progressBar.setValue(0);
            for (int i = 0; i < forestSide; i++) {
                for (int j = 0; j < forestSide; j++) {
                    for (int k = 0; k < grid.grid[i][j].list.size(); k++) {
                        double min = forestSide;
                        for (int s = 0; s < clusterCount; s++) {
                            if (grid.distance(grid.grid[i][j].list.get(k), clusterCenters.list.get(s)) < min) {
                                min = grid.distance(grid.grid[i][j].list.get(k), clusterCenters.list.get(s));
                                grid.grid[i][j].list.get(k).cluster = s;
                            }
                        }
                        clusterList[grid.grid[i][j].list.get(k).cluster].list.add(grid.grid[i][j].list.get(k));
                    }
                }
                progressBar.setValue((100 * i) / forestSide);
                progressBar.paint(progressBar.getGraphics());
            }

            //apply cluster factor to each point in each cluster
            progressBar.setString("Clustering..");
            progressBar.setValue(0);
            int count = 0;
            Random random = new Random();
            for (int i = 0; i < clusterCount; i++) {
                for (int j = 0; j < clusterList[i].list.size(); j++) {
                    count++;
                    //with probability clusterizationFactor, translate tree                    
                    if (random.nextDouble() * 100 < clusterizationRatio) {
                        double increment = 0;
                        boolean tooClose;
                        if (randomCluster) {
                            clusterFactor = random.nextDouble();
                        }
                        do {
                            tooClose = false;
                            EntityPoint tp = grid.translate(clusterList[i].list.get(j), clusterCenters.list.get(i), clusterFactor - increment);
                            
                            if (clusterFactor < 0) {
                                if (grid.getByProximity(tp, minTreeDistance).list.size() > 0) {
                                    tooClose = true;
                                }
                            } else {
                                for (int k = 0; k < j; k++) {
                                    if (grid.distance(tp, clusterList[i].list.get(k)) < minTreeDistance) {
                                        tooClose = true;
                                    }
                                }
                            }
                            if (tooClose) {
                                increment += .05;
                            } else {
                                grid.removePoint(clusterList[i].list.get(j));
                                grid.addPoint(tp);
                                clusterList[i].list.set(j, tp);
                                tp.color = new Color(i % 250, (i * i) % 250, (i * i) % 100);
                            }
                        } while (tooClose);
                    }
                }
                progressBar.setValue((100 * i) / clusterCount);
                progressBar.paint(progressBar.getGraphics());
            }
        } else {
            clusterList = new PointList[1];
            clusterList[0] = grid.getPointList();
        }
        progressBar.setValue(100);
        progressBar.setString("Generating done.");
        progressBar.paint(progressBar.getGraphics());
    }
    
    public Sample sample(String sampleShape, double sampleArea, int sampleCount, String sampleMode, int repetition, JProgressBar progressBar) {
        Sample sample = new Sample(this, sampleShape, sampleArea, sampleCount, sampleMode, repetition);
        sample.plots = null;
        progressBar.setString("Sampling..");
        progressBar.setValue(0);

        //generate circular samples
        if (sampleShape.equals("Circle")) {
            double radius = Math.sqrt(sampleArea / Math.PI);
            if (sampleMode.equals("Random")) {
                sample.grid.generateRandom(sampleCount, 2 * radius, progressBar);
            }
            if (sampleMode.equals("Uniform")) {
                sampleCount = sample.grid.generateUniform(sampleCount);
            }
            PointList sampleCenters = sample.grid.getPointList();
            sample.plots = new PointList[sampleCount];
            
            for (int i = 0; i < sampleCount; i++) {
                sample.plots[i] = grid.getByProximity(sampleCenters.list.get(i), radius);
                progressBar.setValue((100 * i) / sampleCount);
                progressBar.paint(progressBar.getGraphics());
            }
        }

        //generate square samples
        if (sampleShape.equals("Square")) {
            double side = Math.sqrt(sampleArea);
            if (sampleMode.equals("Random")) {
                sample.grid.generateRandom(sampleCount, side, progressBar);
            }
            if (sampleMode.equals("Uniform")) {
                sampleCount = sample.grid.generateUniform(sampleCount);
            }
            PointList sampleCenters = sample.grid.getPointList();
            sample.plots = new PointList[sampleCount];
            
            for (int i = 0; i < sampleCount; i++) {
                sample.plots[i] = grid.getBySquare(sampleCenters.list.get(i), side);
                progressBar.setValue((100 * i) / sampleCount);
                progressBar.paint(progressBar.getGraphics());
            }
        }

        //generate triangle samples
        if (sampleShape.equals("Triangle")) {
            double side = Math.sqrt(sampleArea * 4 / Math.sqrt(3));
            if (sampleMode.equals("Random")) {
                sample.grid.generateRandom(sampleCount, side, progressBar);
            }
            if (sampleMode.equals("Uniform")) {
                sampleCount = sample.grid.generateUniform(sampleCount);
            }
            PointList sampleCenters = sample.grid.getPointList();
            sample.plots = new PointList[sampleCount];
            
            for (int i = 0; i < sampleCount; i++) {
                sample.plots[i] = grid.getByTriangle(sampleCenters.list.get(i), side);
                progressBar.setValue((100 * i) / sampleCount);
                progressBar.paint(progressBar.getGraphics());
            }
        }

        //generate random samples
        if (sampleShape.equals("Random")) {
            double radius = Math.sqrt(sampleArea / Math.PI);
            double triangleSide = Math.sqrt(sampleArea * 4 / Math.sqrt(3));
            double squareSide = Math.sqrt(sampleArea);
            
            if (sampleMode.equals("Random")) {
                sample.grid.generateRandom(sampleCount, triangleSide, progressBar);
            }
            if (sampleMode.equals("Uniform")) {
                sampleCount = sample.grid.generateUniform(sampleCount);
            }
            PointList sampleCenters = sample.grid.getPointList();
            sample.plots = new PointList[sampleCount];
            String[] sampleShapes = new String[sampleCount];
            
            Random random = new Random();
            for (int i = 0; i < sampleCount; i++) {
                int shapeNumber = random.nextInt(3);
                if (shapeNumber == 0) {
                    sample.plots[i] = grid.getByProximity(sampleCenters.list.get(i), radius);
                }
                if (shapeNumber == 1) {
                    sample.plots[i] = grid.getBySquare(sampleCenters.list.get(i), squareSide);
                }
                if (shapeNumber == 2) {
                    sample.plots[i] = grid.getByTriangle(sampleCenters.list.get(i), triangleSide);
                }
                progressBar.setValue((100 * i) / sampleCount);
                progressBar.paint(progressBar.getGraphics());
            }
        }
        progressBar.setValue(100);
        progressBar.setString("Sampling done.");
        progressBar.paint(progressBar.getGraphics());
        
        return sample;
    }
    
    public String saveToFile(File file) {
        try {
            FileOutputStream fileOut = new FileOutputStream(file);
            ObjectOutputStream out = new ObjectOutputStream(fileOut);
            out.writeObject(this);
            out.close();
            fileOut.close();
        } catch (IOException e) {
            e.printStackTrace();
            return "I/O Error!";
        }
        return "Forest succesfuly saved!";
    }
    
    public String loadFromFile(File file) {
        try {
            FileInputStream fileIn =
                    new FileInputStream(file);
            ObjectInputStream in = new ObjectInputStream(fileIn);
            loadFromForest((Forest) in.readObject());
            in.close();
            fileIn.close();
            
        } catch (IOException i) {
            return "I/O error!";
        } catch (ClassNotFoundException c) {
            return "Class error!";
        }
        return "Forest load succesful!";
    }
    
    public void loadFromForest(Forest forest) {
        this.forestArea = forest.forestArea;
        this.tph = forest.tph;
        this.treeCount = forest.treeCount;
        this.minTreeDistance = forest.minTreeDistance;
        this.clusterCount = forest.clusterCount;
        this.clusterFactor = forest.clusterFactor;
        this.clusterRadius = forest.clusterRadius;
        this.clusterizationRatio = forest.clusterizationRatio;
        this.randomCluster = forest.randomCluster;
        this.folding = forest.folding;
        this.forestSide = forest.forestSide;
        this.grid = forest.grid;
        this.clusterGrid = forest.clusterGrid;
        this.clusterList = forest.clusterList;
    }
    
    public void outputToFile(File file, JProgressBar progressBar) {
        progressBar.setString("Writing forest to file..");
        progressBar.setValue(0);
        FileWriter fw = null;
        try {
            if (file == null) {
                file = new File(System.getProperty("user.dir") + "/ForestOutput/" + this.id + ".csv");
            }
            insertForestID();
            boolean fileExists = false;
            if (file.exists()) {
                fileExists = true;
            }
            fw = new FileWriter(file, true);
            BufferedWriter out = new BufferedWriter(fw);
            // if new file, write header
            System.out.println(file.exists());
            
            if (!fileExists) {
                out.write("forestID,X,Y,Weibull,\n");
            }

            //write the forest
            for (int i = 0; i < clusterList.length; i++) {
                
                for (EntityPoint p : clusterList[i].list) {
                    out.write(id + "," + p.x + "," + p.y + "," + p.weibull + "\n");
                }
                progressBar.setValue((100 * i) / clusterList.length);
                progressBar.paint(progressBar.getGraphics());
            }
            out.close();
            progressBar.setValue(100);
            progressBar.setString("Writing forest to file..DONE");
            
        } catch (IOException ex) {
            Logger.getLogger(Forest.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            try {
                fw.close();
            } catch (IOException ex) {
                Logger.getLogger(Forest.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }
    
    public void insertForestID() {
        boolean exists = false;
        try {
            //ForestID exists?
            FileInputStream fstream = new FileInputStream(System.getProperty("user.dir") + "/Registers/forest_register.csv");
            DataInputStream in = new DataInputStream(fstream);
            BufferedReader br = new BufferedReader(new InputStreamReader(in));
            String line;
            while ((line = br.readLine()) != null) {
                if (line.indexOf("" + id) != -1) {
                    exists = true;
                }
            }
            in.close();

            //if does not exist add it 
            FileWriter fw = new FileWriter(System.getProperty("user.dir") + "/Registers/forest_register.csv", true);
            BufferedWriter out = new BufferedWriter(fw);
            if (!exists) {
                out.write(id + "," + tph + "," + clusterFactor + "," + clusterCount + "," + folding + "," + repetition + "\n");
            }
            out.close();
        } catch (Exception e) {//Catch exception if any
            System.err.println("Error: " + e.getMessage());
        }
    }
}